"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Debugger = require("debug");
var methods_1 = require("./methods");
var query_1 = require("./query");
var debug = Debugger('akita:model');
var Model = (function () {
    function Model(data, params) {
        if (data) {
            Object.assign(this, data);
        }
        if (params) {
            Object.defineProperty(this, '__params', {
                value: Object.assign({}, params),
                writable: true,
                enumerable: false,
                configurable: true
            });
        }
    }
    Model.create = function (data) {
        var query = new query_1.default(this, 'create');
        query._data = data || {};
        return query;
    };
    Model.update = function (id, data) {
        var query = new query_1.default(this, 'update');
        if (data) {
            query._data = data;
        }
        if (id && data) {
            query._id = id;
            query._data = data;
        }
        else if (typeof id === 'object') {
            query._data = id;
        }
        return query;
    };
    Model.remove = function (conditions) {
        var query = new query_1.default(this, 'remove');
        if (conditions !== null && typeof conditions === 'object') {
            query.where(conditions);
        }
        else if (conditions !== undefined) {
            query._id = conditions;
        }
        return query;
    };
    Model.count = function (conditions) {
        var query = new query_1.default(this, 'count');
        if (conditions) {
            query.where(conditions);
        }
        return query;
    };
    Model.watch = function (conditions) {
        var query = new query_1.default(this, 'watch');
        if (conditions) {
            query.where(conditions);
        }
        return query;
    };
    Model.find = function (conditions) {
        var query = new query_1.default(this, 'find');
        if (conditions) {
            query.where(conditions);
        }
        return query;
    };
    Model.paginate = function (conditions) {
        var query = new query_1.default(this, 'paginate');
        if (conditions) {
            query.where(conditions);
        }
        return query;
    };
    Model.findOne = function (conditions) {
        var query = new query_1.default(this, 'findOne');
        query.limit(1);
        if (conditions) {
            if (typeof conditions === 'object') {
                query.where(conditions);
            }
            else {
                throw new Error('Akita Error: invalid params for Query#findOne()');
            }
        }
        return query;
    };
    Model.findByPk = function (id) {
        var query = new query_1.default(this, 'findByPk');
        query._id = id;
        return query;
    };
    Model.request = function (path, init, query, reducer) {
        var p = this.path || '';
        if (p[p.length - 1] !== '/' && path) {
            p += '/';
        }
        if (path[0] === '/') {
            path = path.substr(1);
        }
        path = p + path;
        if (path && init && (init.query || init.body)) {
            var queryParams_1 = init.query && Object.assign({}, init.query);
            var bodyParams_1 = init.body && typeof init.body === 'object' && Object.assign({}, init.body);
            var params_1 = {};
            var queryEmits_1 = [];
            var bodyEmits_1 = [];
            path = path.replace(/:\w+/g, function (match) {
                var key = match.substr(1);
                if (queryParams_1 === null || queryParams_1 === void 0 ? void 0 : queryParams_1.hasOwnProperty(key)) {
                    queryEmits_1.push(key);
                    params_1[key] = queryParams_1[key];
                    return encodeURIComponent(queryParams_1[key]);
                }
                if (bodyParams_1 === null || bodyParams_1 === void 0 ? void 0 : bodyParams_1.hasOwnProperty(key)) {
                    bodyEmits_1.push(key);
                    params_1[key] = bodyParams_1[key];
                    return encodeURIComponent(bodyParams_1[key]);
                }
                return match;
            });
            if (query && (queryEmits_1.length || bodyEmits_1.length)) {
                query._params = params_1;
            }
            if (queryParams_1 && queryEmits_1.length) {
                queryEmits_1.forEach(function (key) {
                    delete queryParams_1[key];
                });
                init = Object.assign({}, init, { query: queryParams_1 });
            }
            if (bodyParams_1 === null || bodyParams_1 === void 0 ? void 0 : bodyParams_1.length) {
                bodyEmits_1.forEach(function (key) {
                    delete bodyParams_1[key];
                });
                init = Object.assign({}, init, { body: bodyParams_1 });
            }
        }
        return this.client.request(path, init, query, reducer);
    };
    Model.prototype.request = function (path, init, reducer) {
        var _this = this;
        var M = this.constructor;
        var pk = M.pk || 'id';
        var id = this[pk];
        if (!id) {
            var method = (init === null || init === void 0 ? void 0 : init.method) || 'GET';
            throw new Error("Can not get pk field (" + pk + ") for " + method + ": '" + path + "'");
        }
        id = encodeURIComponent(id);
        if (!path || path[0] === '/') {
            path = id + path;
        }
        else {
            path = id + "/" + path;
        }
        var fullPath = M.path + "/" + path;
        var matchs = fullPath.match(/:\w+/g);
        if (matchs) {
            init = Object.assign({}, init);
            var query_2 = Object.assign({}, init.query);
            matchs.forEach(function (match) {
                var key = match.substr(1);
                if (!query_2.hasOwnProperty(key)) {
                    if (_this.__params && _this.__params.hasOwnProperty(key)) {
                        query_2[key] = _this.__params[key];
                    }
                    else if (_this.hasOwnProperty(key)) {
                        query_2[key] = _this[key];
                    }
                }
            });
            init.query = query_2;
        }
        return M.request(path, init, null, reducer);
    };
    Model.prototype.save = function (init) {
        var _this = this;
        return this.request('', Object.assign({}, {
            method: 'PATCH',
            body: this.toJSON()
        }, init), function (data) {
            if (data && typeof data === 'object') {
                Object.assign(_this, data);
            }
        });
    };
    Model.prototype.remove = function (init) {
        return this.request('', Object.assign({}, {
            method: 'DELETE'
        }, init), function () { });
    };
    Model.prototype.toJSON = function () {
        var json = {};
        for (var key in this) {
            if (this.hasOwnProperty(key)) {
                json[key] = this[key];
            }
        }
        return json;
    };
    return Model;
}());
exports.default = Model;
methods_1.default.forEach(function (method) {
    Model[method] = function (path, init) {
        debug(this.name + "." + method, path, init || '');
        init = init || {};
        init.method = method.toUpperCase();
        return this.request(path, init, null);
    };
    Model.prototype[method] = function (path, init) {
        var M = this.constructor;
        debug(M.name + ".prototype." + method, path, init || '');
        init = init || {};
        init.method = method.toUpperCase();
        return this.request(path, init);
    };
});
